#define MOZ_SOURCE_STAMP 85d16b0be539271da6484435f71c562acd9c3c56
#define MOZ_SOURCE_REPO https://hg.mozilla.org/releases/mozilla-release
#define MOZ_SOURCE_URL https://hg.mozilla.org/releases/mozilla-release/rev/85d16b0be539271da6484435f71c562acd9c3c56
